# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
import requests
import pathlib
import logging
from typing import Tuple

# External imports
import torch
from torch.utils.data import Dataset
import scipy.io
import tqdm
import numpy as np

# Local imports
from .filelist import filelist

SAMPLE_base_link = "https://github.com/benjaminlewis-afrl/SAMPLE_dataset_public/raw/refs/heads/master/mat_files/"


class SAMPLE(Dataset):
    """
    The SAMPLE dataset is made partly from real data provided by MSTAR and partly from synthetic data.

    The dataset is public and will be downloaded if requested and missing on drive.

    It is made of 10 classes of military vehicles: 2s1, bmp2, btr70, m1, m2, m35, m548, m60, t72, zsu23

    Arguments:
        rootdir (str): Path to the root directory where the dataset is stored or will be downloaded
        transform (torchvision.transforms.Compose): A list of torchvision transforms to apply to the complex image
        download (bool): Whether to download the data if missing on disk

    Note:
        An example usage :

        .. code-block:: python

            import torchcvnn
            from torchcvnn.datasets import SAMPLE

            transform = v2.Compose(
                transforms=[v2.ToImage(), v2.Resize(128), v2.CenterCrop(128)]
            )
            dataset = SAMPLE(
                rootdir, transform=transform, download=True
            )
            X, y = dataset[0]

        Displayed below are some examples drawn randomly from SAMPLE. To plot them, we extracted
        only the magnitude of the signals although the data are indeed complex valued.

        .. image:: ../assets/datasets/SAMPLE.png
           :alt: Samples from MSTAR
           :width: 60%

    """

    def __init__(self, rootdir: str, transform=None, download: bool = False):
        super().__init__()
        self.rootdir = pathlib.Path(rootdir)
        self.transform = transform

        self.class_names = list(filelist["real"].keys())

        # We look into rootdir if the data are available
        self.tot_num_samples = 0
        for cl in self.class_names:
            self.tot_num_samples += len(filelist["real"][cl])
            self.tot_num_samples += len(filelist["synth"][cl])

        self.data_files = {}
        self.num_data_files = {}
        self.tot_num_data_files = 0

        pbar = tqdm.tqdm(total=self.tot_num_samples)
        for cl in self.class_names:
            for mode in ["real", "synth"]:
                for filename in filelist[mode][cl]:
                    filepath = self.rootdir / cl / filename
                    if not filepath.exists():
                        if download:
                            url = f"{SAMPLE_base_link}{mode}/{cl}/{filename}"
                            self.download_file(url, filepath)
                        else:
                            raise FileNotFoundError(f"{filepath} not found")
                    pbar.update(1)
            self.data_files[cl] = list((self.rootdir / cl).glob("*.mat"))
            self.num_data_files[cl] = len(self.data_files[cl])
            self.tot_num_data_files += self.num_data_files[cl]

    def download_file(self, url: str, filepath: pathlib.Path):
        """
        Download a file from an URL and save it on disk

        Args:
            url (str): URL to download the file from
            filepath (pathlib.Path): Path to save the file
        """
        logging.debug(f"Downloading {url} to {filepath}")
        # Ensure the target directory exists
        filepath.parent.mkdir(parents=True, exist_ok=True)

        # Donwload and save the file on disk
        response = requests.get(url)
        with open(filepath, "wb") as fh:
            fh.write(response.content)

    def __len__(self) -> int:
        """
        Return the total number of samples in the dataset
        """
        return self.tot_num_data_files

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, int]:
        """
        Return a sample from the dataset
        """
        if index >= self.tot_num_data_files:
            raise IndexError

        # We look for the class from which the sample will be taken
        for key in self.data_files.keys():
            if index < self.num_data_files[key]:
                break
            index -= self.num_data_files[key]

        filename = self.data_files[key][index]
        logging.debug(f"Reading SAMPLE file : {filename}")

        data = scipy.io.loadmat(filename)

        # Below are the keys available in the mat files
        # dict_keys(['__header__', '__version__', '__globals__', 'aligned', 'azimuth', 'bandwidth', 'center_freq', 'complex_img', 'complex_img_unshifted', 'elevation', 'explanation', 'range_pixel_spacing', 'range_resolution', 'source_mstar_file', 'target_name', 'taylor_weights', 'xrange_pixel_spacing', 'xrange_resolution']

        meta = {
            k: data[k]
            for k in [
                "azimuth",
                "elevation",
                "bandwidth",
                "center_freq",
                "range_pixel_spacing",
                "range_resolution",
                "xrange_pixel_spacing",
                "xrange_resolution",
            ]
        }

        complex_img = data["complex_img"][:, :, np.newaxis]

        class_idx = self.class_names.index(key)

        if self.transform is not None:
            complex_img = self.transform(complex_img)

        return complex_img, class_idx, meta



# coding: utf-8

# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
from enum import Enum
import pathlib
import logging
from typing import Union

# External imports
import torch
from torch.utils.data import Dataset
import h5py  # Required because the data are matlab v7.3 files
import numpy as np


class CINEView(Enum):
    SAX = 1
    LAX = 2


class AccFactor(Enum):
    ACC4 = 4
    ACC8 = 8
    ACC10 = 10


def load_matlab_file(filename: str, key: str) -> np.ndarray:
    """
    Load a matlab file in HDF5 format
    """
    with h5py.File(filename, "r") as f:
        logging.debug(f"Got the keys {f.keys()} from {filename}")
        data = f[key][()]
    return data


def kspace_to_image(kspace: Union[torch.Tensor, np.ndarray]) -> torch.Tensor:
    """
    Convert k-space data to image data. The returned kspace is
    of the same type than the the provided image (np.ndarray or torch.Tensor).

    Arguments:
        kspace : torch.Tensor or np.ndarray
            k-space data

    Returns:
        torch.Tensor or np.ndarray
            image data
    """
    if isinstance(kspace, torch.Tensor):
        img = torch.fft.fftshift(torch.fft.ifft2(torch.fft.ifftshift(kspace)))
    else:
        img = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(kspace)))
    return img


def image_to_kspace(
    img: Union[torch.Tensor, np.ndarray]
) -> Union[torch.Tensor, np.ndarray]:
    """
    Convert image data to k-space data. The returned kspace is
    of the same type than the the provided image (np.ndarray or torch.Tensor)

    Arguments:
        img : torch.Tensor or np.ndarray
            Image data

    Returns:
        torch.Tensor or np.ndarray
            k-space data

    """
    if isinstance(img, torch.Tensor):
        kspace = torch.fft.fftshift(torch.fft.fft2(torch.fft.ifftshift(img)))
    else:
        kspace = np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(img)))
    return kspace


def combine_coils_from_kspace(kspace: np.ndarray) -> np.ndarray:
    """
    Combine the coils of the k-space data using the root sum of squares

    Arguments:
        kspace : np.ndarray
            k-space data of shape (sc, ky, kx)

    Returns:
        np.ndarray
            Image data with coils combined, of shape (ky, kx), real valued, positive
    """
    if kspace.ndim != 3:
        raise ValueError(
            f"kspace should have 3 dimensions, got {kspace.ndim}. Expected dimensions (sc, ky, kx)"
        )
    images = np.fft.ifft2(np.fft.ifftshift(kspace))
    return np.fft.fftshift(np.sqrt(np.sum(np.abs(images) ** 2, axis=0)))


class MICCAI2023(Dataset):
    """
    Loads the MICCAI2023 challenge data for the reconstruction task Task 1

    The data are described on https://cmrxrecon.github.io/Task1-Cine-reconstruction.html

    You need to download the data before hand in order to use this class.

    For loading the data, you may want to alternatively consider the fastmri library, see https://github.com/facebookresearch/fastMRI/

    The structure of the dataset is as follows:

        rootdir/ChallengeData/MultiCoil/cine/TrainingSet/P{id}/
                                    - cine_sax.mat
                                    - cin_lax.mat
        rootdir/ChallengeData/MultiCoil/cine/TrainingSet/AccFactor04/P{id}/
                                    - cine_sax.mat
                                    - cine_sax_mask.mat
                                    - cin_lax.mat
                                    - cine_lax_mask.mat
        rootdir/ChallengeData/MultiCoil/cine/TrainingSet/AccFactor08/P{id}/
                                    - cine_sax.mat
                                    - cine_sax_mask.mat
                                    - cin_lax.mat
                                    - cine_lax_mask.mat
        rootdir/ChallengeData/MultiCoil/cine/TrainingSet/AccFactor10/P{id}/
                                    - cine_sax.mat
                                    - cine_sax_mask.mat
                                    - cin_lax.mat
                                    - cine_lax_mask.mat

    The cine_sax or sine_lax files are :math:`(k_x, k_y, s_c, s_z, t)` where :

    - :math:`k_x`: matrix size in x-axis (k-space)
    - :math:`k_y``: matrix size in y-axis (k-space)
    - :math:`s_c`: coil array number (compressed to 10)
    - :math:`s_x`: matrix size in x-axis (image)
    - :math:`s_y`: matrix size in y-axis (image) , used in single-coil data
    - :math:`s_z`: slice number for short axis view, or slice group for long axis (i.e., 3ch, 2ch and 4ch views)
    - :math:`t`: time frame.

    Note the k-space dimensions (in x/y axis) are not the same depending on the patient.

    This is a recontruction dataset. The goal is to reconstruct the fully sampled k-space
    from the subsampled k-space. The acceleratation factor specifies the subsampling rate.

    There are also the Single-Coil data which is not yet considered by this implementation

    Note:
        An example usage :

        .. code-block:: python

            import torchcvnn
            from torchcvnn.datasets.miccai2023 import MICCAI2023, CINEView, AccFactor

            def process_kspace(kspace, coil_idx, slice_idx, frame_idx):
                coil_kspace = kspace[:, :, coil_idx, slice_idx, frame_idx]
                mod_kspace = np.log(np.abs(coil_kspace) + 1e-9)

                img = kspace_to_image(coil_kspace)
                img = np.abs(img)
                img = img / img.max()

                return mod_kspace, img

            dataset = MICCAI2023(rootdir, view=CINEView.SAX, acc_factor=AccFactor.ACC8)
            subsampled_kspace, subsampled_mask, full_kspace = dataset[0]

            frame_idx = 5
            slice_idx = 0
            coil_idx = 9

            mod_full, img_full = process_kspace(full_kspace, coil_idx, slice_idx, frame_idx)
            mod_sub, img_sub = process_kspace(subsampled_kspace, coil_idx, slice_idx, frame_idx)

            # Plot the above magnitudes
            ...


        Displayed below is an example patient with the SAX view and acceleration of 8:

        .. figure:: ../assets/datasets/miccai2023_sax8.png
           :alt: Example patient from the MICCAI2023 dataset with both the full sampled and under sampled k-space and images
           :width: 100%
           :align: center

        Displayed below is an example patient with the LAX view and acceleration of 4:

        .. figure:: ../assets/datasets/miccai2023_lax4.png
           :alt: Example patient from the MICCAI2023 dataset with both the full sampled and under sampled k-space and images
           :width: 100%
           :align: center

        You can combine the coils using the root sum of squares
        to get a magnitude image (real valued) with all the
        coil contributions.


        Below are examples combining the coils for a given
        frame and slice, for LAX (top) and SAX (bottom). It uses
        the function :py:func:`torchcvnn.datasets.miccai2023.combine_coils_from_kspace`

        .. figure:: ../assets/datasets/miccai2023_combined_lax.png
           :alt: Example LAX, combining the coils
           :width: 50%
           :align: center

        .. figure:: ../assets/datasets/miccai2023_combined_sax.png
           :alt: Example SAX, combining the coils
           :width: 50%
           :align: center

    """

    def __init__(
        self,
        rootdir: str,
        view: CINEView = CINEView.SAX,
        acc_factor: AccFactor = AccFactor.ACC4,
    ):
        self.rootdir = pathlib.Path(rootdir)

        if view == CINEView.SAX:
            self.input_filename = "cine_sax.mat"
            self.mask_filename = "cine_sax_mask.mat"
        elif view == CINEView.LAX:
            self.input_filename = "cine_lax.mat"
            self.mask_filename = "cine_lax_mask.mat"

        # List all the available data
        self.fullsampled_rootdir = self.rootdir / "MultiCoil" / "cine" / "TrainingSet"
        self.fullsampled_key = "kspace_full"
        self.subsampled_rootdir = (
            self.rootdir
            / "MultiCoil"
            / "cine"
            / "TrainingSet"
            / f"AccFactor{acc_factor.value:02d}"
        )
        self.subsampled_key = f"kspace_sub{acc_factor.value:02d}"
        self.mask_key = f"mask{acc_factor.value:02d}"

        logging.info(f"Loading data from {self.subsampled_rootdir}")

        # We list all the patients in the subsampled data directory
        # and check we have the data, mask and full sampled data
        self.patients = []
        for patient in self.subsampled_rootdir.iterdir():
            if not patient.is_dir():
                continue

            if not (patient / self.input_filename).exists():
                logging.warning(f"Missing {self.input_filename} for patient {patient}")
                continue

            if not (patient / self.mask_filename).exists():
                logging.warning(f"Missing {self.mask_filename} for patient {patient}")
                continue

            fullsampled_patient = self.fullsampled_rootdir / patient.name
            if not (fullsampled_patient / self.input_filename).exists():
                logging.warning(
                    f"Missing {self.input_filename} for patient {fullsampled_patient}"
                )
                continue

            self.patients.append(patient)

        logging.debug(
            f"I found {len(self.patients)} patient(s) : {[p.name for p in self.patients]}"
        )

    def __len__(self):
        return len(self.patients)

    def __getitem__(self, idx):
        """
        Returns the subsampled k-space data, the mask and the fully sampled k-space data
        """
        patient = self.patients[idx]

        subsampled_data = None
        subsampled_mask = None
        fullsampled_data = None

        # Load the subsampled data
        logging.info(f"Loading {patient / self.input_filename}")
        subsampled_data = load_matlab_file(
            patient / self.input_filename, self.subsampled_key
        ).transpose(3, 4, 2, 1, 0)
        subsampled_data = subsampled_data["real"] + 1j * subsampled_data["imag"]
        # (kx, ky, sc, sz, t) for multi-coil data
        # e.g. (246, 512, 10, 10, 12)

        logging.info(f"Loading {patient / self.mask_filename}")
        subsampled_mask = load_matlab_file(
            patient / self.mask_filename, self.mask_key
        ).transpose(0, 1)
        # (kx, ky)
        # e.g. (246, 512)

        logging.info(
            f"Loading {self.fullsampled_rootdir / patient.name / self.input_filename}"
        )
        fullsampled_data = load_matlab_file(
            self.fullsampled_rootdir / patient.name / self.input_filename,
            self.fullsampled_key,
        ).transpose(3, 4, 2, 1, 0)
        fullsampled_data = fullsampled_data["real"] + 1j * fullsampled_data["imag"]
        # kx, ky, sc, sz, t
        # e.g. (246, 512, 10, 10, 12)

        return subsampled_data, subsampled_mask, fullsampled_data



# MIT License

# Copyright (c) 2023 Jérémie Levi, Victor Dhédin, Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
import time

# External imports
import torch

# Local imports
import torchcvnn.nn as c_nn
import torchcvnn.nn.modules.batchnorm as bn


def test_inv():
    # Take a positive semi-definite matrix
    B = 10
    X = torch.randn((B, 2, 2))
    XT = torch.transpose(X, 1, 2)
    XXT = torch.bmm(X, XT) + (0.01 * torch.eye(2)).unsqueeze(0)

    # Compute its inverse square root
    inv = bn.inv_2x2(XXT)
    torch_inv = torch.linalg.inv(XXT)

    # And check this is really the inerse of X
    assert torch.allclose(inv, torch_inv)


def test_inv_sqrt():
    # Take a positive semi-definite matrix
    B = 10
    X = torch.randn((B, 2, 2))
    XT = torch.transpose(X, 1, 2)
    XXT = torch.bmm(X, XT) + (0.05 * torch.eye(2)).unsqueeze(0)

    # Compute its inverse square root
    inv_sqrt = bn.inv_sqrt_2x2(XXT)

    # Square it to have the inverse
    inv = torch.bmm(inv_sqrt, inv_sqrt)

    torch_inv = torch.linalg.inv(XXT)

    # And check this is really the inerse of X
    assert torch.allclose(inv, torch_inv)


def time_it(f, N=1000):
    t0 = time.time()
    for i in range(N):
        f()
    t1 = time.time()
    return (t1 - t0) / N


def time_inv_sqrt():
    # Take a positive semi-definite matrix
    B = 10
    X = torch.randn((B, 2, 2))
    XT = torch.transpose(X, 1, 2)
    XXT = torch.bmm(X, XT) + (0.05 * torch.eye(2)).unsqueeze(0)

    # Compute its inverse square root
    el1 = time_it(lambda: bn.slow_inv_sqrt_2x2(XXT))
    el2 = time_it(lambda: bn.inv_sqrt_2x2(XXT))
    print(el1, el2)


def test_batchnorm1d():
    B, C = 20, 16
    m = c_nn.BatchNorm1d(C)

    x = torch.randn((B, C), dtype=torch.complex64)
    output = m(x)

    # Compute the variance/covariance
    xc = output.transpose(0, 1).reshape(C, -1)  # C, B
    mus = xc.mean(axis=-1)  # 16 means
    xc_real = torch.view_as_real(xc)
    covs = bn.batch_cov(xc_real)  # 16 covariances matrices

    # All the mus must be 0's
    # For some reasons, this is not exactly 0
    assert torch.allclose(mus, torch.zeros_like(mus), atol=1e-7)
    # All the covs must be identities
    id_cov = 0.5 * torch.eye(2).tile(C, 1, 1)
    assert torch.allclose(covs, id_cov, atol=1e-3)


def test_batchnorm2d():
    B, C, H, W = 20, 16, 50, 100
    m = c_nn.BatchNorm2d(C)

    x = torch.randn((B, C, H, W), dtype=torch.complex64)
    output = m(x)

    # Compute the variance/covariance
    xc = output.transpose(0, 1).reshape(C, -1)  # C, BxHxW
    mus = xc.mean(axis=-1)  # 16 means
    xc_real = torch.view_as_real(xc)
    covs = bn.batch_cov(xc_real)  # 16 covariances matrices

    # All the mus must be 0's
    # For some reasons, this is not exactly 0
    assert torch.allclose(mus, torch.zeros_like(mus), atol=1e-7)
    # All the covs must be identities
    id_cov = 0.5 * torch.eye(2).tile(C, 1, 1)
    assert torch.allclose(covs, id_cov, atol=1e-3)


if __name__ == "__main__":
    test_inv()
    test_inv_sqrt()
    # time_inv_sqrt()
    test_batchnorm1d()
    test_batchnorm2d()



# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
import time

# External imports
import torch

# Local imports
import torchcvnn.nn as c_nn
import torchcvnn.nn.modules.batchnorm as bn


def test_layernorm_real():
    # NLP Example
    batch, sentence_length, embedding_dim = 20, 5, 10
    embedding = torch.randn(batch, sentence_length, embedding_dim)
    layer_norm = torch.nn.LayerNorm(embedding_dim)
    layer_norm.eval()
    # Activate module
    out = layer_norm(embedding)

    assert out.shape == embedding.shape

    out = out.view(batch, sentence_length, -1)  # batch, sentence_length, embeddig_dim

    mus = out.mean(axis=-1)
    assert torch.allclose(mus, torch.zeros_like(mus), atol=1e-6)
    stds = out.std(axis=-1)
    assert torch.allclose(stds, torch.ones_like(stds), atol=1e-1)

    # Image Example
    N, C, H, W = 20, 5, 10, 10
    x = torch.randn(N, C, H, W)
    # Normalize over the last three dimensions (i.e. the channel and spatial dimensions)
    # as shown in the image below
    layer_norm = torch.nn.LayerNorm([C, H, W])
    layer_norm.eval()
    out = layer_norm(x)

    out = out.view(N, -1)  # N, C*H*W

    mus = out.mean(axis=-1)
    assert torch.allclose(mus, torch.zeros_like(mus), atol=1e-6)
    stds = out.std(axis=-1)
    assert torch.allclose(stds, torch.ones_like(stds), atol=1e-1)


def test_layernorm():
    #############
    # NLP Example
    batch, sentence_length, embedding_dim = 20, 5, 10
    embedding = torch.randn(
        (batch, sentence_length, embedding_dim), dtype=torch.complex64
    )
    layer_norm = c_nn.LayerNorm(embedding_dim)
    layer_norm.eval()
    # For the unit test, we randomly initialize the bias
    # to avoid testing only the particular case bias = 0
    torch.nn.init.normal_(layer_norm.bias)

    # Activate module
    out = layer_norm(embedding)

    assert out.shape == embedding.shape  # batch x sentence_length x embedding_dim

    # The out tensor is considered as a collection of batch x sentence_length
    # vectors of size embedding_dim
    out = out.view(-1, embedding_dim).transpose(0, 1)

    mus = out.mean(axis=-1)
    # Check the means equal zeros. Note that, in general, we expect
    assert torch.allclose(mus, layer_norm.bias, atol=1e-6)

    out_real = torch.view_as_real(out)
    covs = bn.batch_cov(out_real)
    id_cov = 0.5 * torch.eye(2).tile(embedding_dim, 1, 1)

    assert torch.allclose(covs, id_cov, atol=1e-3)

    ###############
    # Image Example
    N, C, H, W = 20, 5, 10, 10
    embedding_dim = C * H * W
    x = torch.randn((N, C, H, W), dtype=torch.complex64)

    # Normalize over the last three dimensions (i.e. the channel and spatial dimensions)
    # as shown in the image below
    layer_norm = c_nn.LayerNorm([C, H, W])
    layer_norm.eval()
    # For the unit test, we randomly initialize the bias
    # to avoid testing only the particular case bias = 0
    torch.nn.init.normal_(layer_norm.bias)

    # Activate module
    out = layer_norm(x)

    assert out.shape == x.shape  # N, C, H, W

    # The out tensor is considered as a collection of N
    # vectors of size embedding_dim
    out = out.view(-1, embedding_dim).transpose(0, 1)

    mus = out.mean(axis=-1)
    # Check the means equal zeros. Note that, in general, we expect
    assert torch.allclose(mus, layer_norm.bias, atol=1e-6)

    out_real = torch.view_as_real(out)
    covs = bn.batch_cov(out_real)
    id_cov = 0.5 * torch.eye(2).tile(embedding_dim, 1, 1)

    assert torch.allclose(covs, id_cov, atol=1e-3)


def test_rmsnorm():
    #############
    # NLP Example
    batch, sentence_length, embedding_dim = 20, 5, 10
    embedding = torch.randn(
        (batch, sentence_length, embedding_dim), dtype=torch.complex64
    )
    rms_norm = c_nn.RMSNorm(embedding_dim)
    rms_norm.eval()

    # Activate module
    out = rms_norm(embedding)

    assert out.shape == embedding.shape  # batch x sentence_length x embedding_dim

    ###############
    # Image Example
    N, C, H, W = 20, 5, 10, 10
    embedding_dim = C * H * W
    x = torch.randn((N, C, H, W), dtype=torch.complex64)

    # Normalize over the last three dimensions (i.e. the channel and spatial dimensions)
    # as shown in the image below
    rms_norm = c_nn.RMSNorm([C, H, W])
    rms_norm.eval()

    # Activate module
    out = rms_norm(x)

    assert out.shape == x.shape  # N, C, H, W


if __name__ == "__main__":
    test_layernorm_real()
    test_layernorm()
    test_rmsnorm()



# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
import os
import pathlib

# External imports
import numpy as np

# Local imports
from .ann_file import AnnFile


def parse_slc_filename(filename):
    """
    Parses a filename of a SLC file
    {site name}_{line ID}_{flight ID}_{data take counter}_{acquisition date}_{band}{steering}{polarization}_{stack_version}... _{baseline correction}_{segment number}_{downsample factor}.slc

    and returns all the information in a dictionary
    """
    # Remove the .slc extension and split the fields
    fields = filename[:-4].split("_")
    parameters = {
        "site_name": fields[0],
        "line_ID": fields[1],
        "flight_ID": fields[2],
        "data_take_counter": fields[3],
        "acquisition_date": fields[4],
        "band": fields[5][0],
        "steering": fields[5][1:-2],
        "polarization": fields[5][-2:],
        "stack_version": fields[6],
        "baseline_correction": fields[7],
        "segment_number": int(
            fields[8][1:]
        ),  # the segment is encoded as  s{segment_number}
        "downsample_factor": fields[9],
    }
    return parameters


class SLCFile:
    r"""
    Reads a SLC file

    The filenames contain interesting information:

    {site name}_{line ID}_{flight ID}_{data take counter}_{acquisition date}_{band}{steering}{polarization}_{stack_version}... _{baseline correction}_{segment number}_{downsample factor}.slc

    e.g. SSurge_15305_14170_007_141120_L090HH_01_BC_s1_1x1.slc is

    - site_name : SSurge
    - line ID : 15305
    - flight ID : 14170
    - data take counter : 007
    - acquisition date : 141120, the date is in YYMMDD format (UTC time).
    - band : L
    - steering : 090
    - polarization : HH
    - stack version : 01
    - baseline correction : BC, means the data is corrected for residual baseline
    - segment number : s1
    - downsample factor : 1x1

    There is one SLC file per segment and per polarization.
    """

    def __init__(self, filename: str, patch_size: tuple, patch_stride: tuple = None):
        self.filename = pathlib.Path(filename)
        self.parameters = parse_slc_filename(self.filename.name)
        self.patch_size = patch_size
        self.patch_stride = patch_stride
        if self.patch_stride is None:
            self.patch_stride = patch_size

        # The annotation filename is almost the same as the SLC filename, except we drop
        # the segment number and downsample factor
        # We expect it to be colocated with the SLC file
        ann_filename = "_".join(str(self.filename.name).split("_")[:-2]) + ".ann"
        self.ann_file = AnnFile(str(self.filename.parent / ann_filename))

        downsample_factor = self.parameters["downsample_factor"]
        segment_number = self.parameters["segment_number"]
        # self.azimuth_pixel_spacing = getattr(
        #     self.ann_file, f"{downsample_factor}_slc_azimuth_pixel_spacing"
        # )
        # self.range_pixel_spacing = getattr(
        #     self.ann_file, f"{downsample_factor}_slc_range_pixel_spacing"
        # )
        # self.global_average_squint_angle = self.ann_file.global_average_squint_angle
        # self.center_wavelength = self.ann_file.center_wavelength
        self.n_rows = getattr(
            self.ann_file, f"slc_{segment_number}_{downsample_factor}_rows"
        )
        self.n_cols = getattr(
            self.ann_file, f"slc_{segment_number}_{downsample_factor}_columns"
        )

        # Precompute the dimension of the grid of patches
        nrows_patch, ncols_patch = self.patch_size
        row_stride, col_stride = self.patch_stride

        self.nsamples_per_rows = (self.n_rows - nrows_patch) // row_stride + 1
        self.nsamples_per_cols = (self.n_cols - ncols_patch) // col_stride + 1

    @property
    def key(self):
        return "_".join(
            (
                self.parameters["site_name"],
                self.parameters["line_ID"],
                self.parameters["flight_ID"],
                self.parameters["data_take_counter"],
                self.parameters["acquisition_date"],
                self.parameters["band"],
                self.parameters["steering"],
                self.parameters["stack_version"],
                self.parameters["baseline_correction"],
            )
        )

    @property
    def polarization(self):
        return self.parameters["polarization"]

    def __len__(self):
        """
        Returns the number of patches that can be extracted from the SLC file
        """
        return self.nsamples_per_rows * self.nsamples_per_cols

    def __getitem__(self, item):
        """
        Returns the item-th patch from the SLC file
        """
        assert 0 <= item < len(self)
        # Compute the row and column index of the patch
        row = item // self.nsamples_per_cols
        col = item % self.nsamples_per_cols

        # Compute the starting row and column index of the patch
        row_stride, col_stride = self.patch_stride
        row_start = row * row_stride
        col_start = col * col_stride

        # Read the patch
        # The SLC file is a binary file of complex64
        patch = np.zeros(self.patch_size, dtype=np.complex64)
        with open(self.filename, "rb") as fh:
            for row in range(self.patch_size[0]):
                fh.seek(
                    (row_start + row) * self.n_cols * 8 + col_start * 8, os.SEEK_SET
                )
                patch[row, :] = np.fromfile(
                    fh, dtype=np.complex64, count=self.patch_size[1]
                )
        return patch



# MIT License

# Copyright (c) 2024-2025 Jeremy Fix, Huy Nguyen

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# External imports
import torch
import numpy as np

# Local imports
import torchcvnn.transforms as transforms


def test_fft_resize_ndarray():
    # Create a random complex tensor
    tensor = np.random.rand(100, 100) + 1j * np.random.rand(100, 100)

    # Resize the tensor
    target_size = (59, 49)
    spatial_resize = transforms.FFTResize(target_size)
    resized_tensor = spatial_resize(tensor)

    assert (resized_tensor.shape[-2], resized_tensor.shape[-1]) == target_size
    assert type(resized_tensor) == np.ndarray
    assert resized_tensor.dtype in [np.complex64, np.complex128]

    # Resize the tensor
    target_size = (123, 121)
    spatial_resize = transforms.FFTResize(target_size)
    resized_tensor = spatial_resize(tensor)
    assert (resized_tensor.shape[-2], resized_tensor.shape[-1]) == target_size
    assert type(resized_tensor) == np.ndarray
    assert resized_tensor.dtype in [np.complex64, np.complex128]


def test_fft_resize_tensor():
    # Create a random complex tensor
    tensor = np.random.rand(100, 100) + 1j * np.random.rand(100, 100)
    tensor = torch.as_tensor(tensor)

    # Resize the tensor
    target_size = (59, 49)
    spatial_resize = transforms.FFTResize(target_size)
    resized_tensor = spatial_resize(tensor)

    assert (resized_tensor.shape[-2], resized_tensor.shape[-1]) == target_size
    assert type(resized_tensor) == torch.Tensor
    assert resized_tensor.dtype in [torch.complex64, torch.complex128]

    # Resize the tensor
    target_size = (123, 121)
    spatial_resize = transforms.FFTResize(target_size)
    resized_tensor = spatial_resize(tensor)
    assert (resized_tensor.shape[-2], resized_tensor.shape[-1]) == target_size
    assert type(resized_tensor) == torch.Tensor
    assert resized_tensor.dtype in [torch.complex64, torch.complex128]


def test_spatial_resize_ndarray():
    # Create a random complex tensor
    tensor = np.random.rand(100, 100) + 1j * np.random.rand(100, 100)

    # Resize the tensor
    target_size = (59, 48)
    spatial_resize = transforms.SpatialResize(target_size)
    resized_tensor = spatial_resize(tensor)

    assert resized_tensor.shape == target_size
    assert type(resized_tensor) == np.ndarray
    assert resized_tensor.dtype == np.complex64

    # Resize the tensor
    target_size = (123, 121)
    spatial_resize = transforms.SpatialResize(target_size)
    resized_tensor = spatial_resize(tensor)

    assert resized_tensor.shape == target_size
    assert type(resized_tensor) == np.ndarray
    assert resized_tensor.dtype == np.complex64


def test_spatial_resize_tensor():
    # Create a random complex tensor
    tensor = np.random.rand(100, 100) + 1j * np.random.rand(100, 100)
    tensor = torch.as_tensor(tensor)

    # Resize the tensor
    target_size = (50, 50)
    spatial_resize = transforms.SpatialResize(target_size)
    resized_tensor = spatial_resize(tensor)

    assert resized_tensor.shape == target_size
    assert type(resized_tensor) == torch.Tensor
    assert resized_tensor.dtype == torch.complex64

    # Resize the tensor
    target_size = (121, 121)
    spatial_resize = transforms.SpatialResize(target_size)
    resized_tensor = spatial_resize(tensor)

    assert resized_tensor.shape == target_size
    assert type(resized_tensor) == torch.Tensor
    assert resized_tensor.dtype == torch.complex64


def on_rgb_img():
    """
    Test the resize from an image
    """
    from PIL import Image
    import matplotlib.pyplot as plt

    im = Image.open("img_rgb.jpg")

    array = np.array(im)
    array = array - 1j * array
    array = np.transpose(array, (2, 0, 1)) / 255.0

    # 487, 565
    target_size = (123, 456)

    spatial_resize = transforms.SpatialResize(target_size)
    resized1 = spatial_resize(array)

    fft_resize = transforms.FFTResize(target_size)
    resized2 = fft_resize(array)

    plt.subplot(1, 3, 1)
    plt.imshow(array.real.transpose(1, 2, 0), clim=(0, 1))
    plt.title("Original")
    plt.subplot(1, 3, 2)
    plt.imshow(resized1.real.transpose(1, 2, 0), clim=(0, 1))
    plt.title("Spatial resize")
    plt.subplot(1, 3, 3)
    plt.imshow(resized2.real.transpose(1, 2, 0), clim=(0, 1))
    plt.title("FFT resize")
    plt.show()


def on_bw_img():
    """
    Test the resize from an image
    """
    from PIL import Image
    import matplotlib.pyplot as plt

    im = Image.open("img_bw.jpg")

    array = np.array(im)
    array = array - 1j * array
    array = array / 255.0

    target_size = (123, 456)

    spatial_resize = transforms.SpatialResize(target_size)
    resized1 = spatial_resize(array)

    fft_resize = transforms.FFTResize(target_size)
    resized2 = fft_resize(array)

    plt.subplot(1, 3, 1)
    plt.imshow(array.real, cmap="gray", clim=(0, 1))
    plt.title("Original")
    plt.subplot(1, 3, 2)
    plt.imshow(resized1.real, cmap="gray", clim=(0, 1))
    plt.title("Spatial resize")
    plt.subplot(1, 3, 3)
    plt.imshow(resized2.real, cmap="gray", clim=(0, 1))
    plt.title("FFT resize")
    plt.show()


if __name__ == "__main__":
    test_fft_resize_ndarray()
    test_fft_resize_tensor()
    test_spatial_resize_ndarray()
    test_spatial_resize_tensor()
    on_rgb_img()
    on_bw_img()



# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
import requests
from bs4 import BeautifulSoup
import re

SAMPLE_base_link = "https://github.com/benjaminlewis-afrl/SAMPLE_dataset_public/blob/refs/heads/master/mat_files/"

SAMPLE_classes = [
    "2s1",
    "bmp2",
    "btr70",
    "m1",
    "m2",
    "m35",
    "m548",
    "m60",
    "t72",
    "zsu23",
]


def generate_file_list():
    """
    Download the list of mat files available for every mode and every class
    and save them as a dictionnary to be loaded by torchcvnn

    We do this "asynchronously" because it happens, from time to time, that
    requests is returning an empty list of files. Requesting and parsing is anyway
    some overhead we can avoid at runtime.
    """
    with open("filelist.py", "w") as fh:
        fh.write("filelist = {\n")
        for mode in ["real", "synth"]:
            fh.write(f'"{mode}": {{\n')
            for cl in SAMPLE_classes:
                url = f"{SAMPLE_base_link}{mode}/{cl}/"
                result = requests.get(url)
                soup = BeautifulSoup(result.text, "html.parser")
                matfiles = soup.find_all(title=re.compile("\.mat$"))

                fh.write(f'"{cl}": [\n')
                fh.write(",\n".join([f'\t"{f.get("title")}"' for f in matfiles]))
                fh.write("],\n")

                if len(matfiles) == 0:
                    # print("=====================================\n")
                    print(f"Empty list of files for {cl} in {mode}")
                    # print(result.text)
                    print(
                        "============================================================\n"
                    )

            fh.write("},\n")
        fh.write("}\n")


if __name__ == "__main__":
    generate_file_list()



# MIT License

# Copyright (c) 2025 Rodolphe Durand

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Most of the code in this file is from :
# https://code.activestate.com/recipes/410469-xml-as-dictionary/

from typing import Any
from xml.etree import ElementTree


class XmlListConfig(list):
    def __init__(self, aList):
        for element in aList:
            if element:
                # treat like dict
                if len(element) == 1 or element[0].tag != element[1].tag:
                    self.append(XmlDictConfig(element))
                # treat like list
                elif element[0].tag == element[1].tag:
                    self.append(XmlListConfig(element))
            elif element.text:
                text = element.text.strip()
                if text:
                    self.append(text)


class XmlDictConfig(dict):
    """
    Example usage:

    >>> tree = ElementTree.parse('your_file.xml')
    >>> root = tree.getroot()
    >>> xmldict = XmlDictConfig(root)

    Or, if you want to use an XML string:

    >>> root = ElementTree.XML(xml_string)
    >>> xmldict = XmlDictConfig(root)

    And then use xmldict for what it is... a dict.
    """

    def __init__(self, parent_element):
        if parent_element.items():
            self.update(dict(parent_element.items()))
        for element in parent_element:
            if element:
                # treat like dict - we assume that if the first two tags
                # in a series are different, then they are all different.
                if len(element) == 1 or element[0].tag != element[1].tag:
                    aDict = XmlDictConfig(element)
                # treat like list - we assume that if the first two tags
                # in a series are the same, then the rest are the same.
                else:
                    # here, we put the list in dictionary; the key is the
                    # tag name the list elements all share in common, and
                    # the value is the list itself
                    aDict = {element[0].tag: XmlListConfig(element)}
                # if the tag has attributes, add those to the dict
                if element.items():
                    aDict.update(dict(element.items()))
                self.update({element.tag: aDict})
            # this assumes that if you've got an attribute in a tag,
            # you won't be having any text. This may or may not be a
            # good idea -- time will tell. It works for the way we are
            # currently doing XML configuration files...
            elif element.items():
                self.update({element.tag: dict(element.items())})
            # finally, if there are no child tags and no attributes, extract
            # the text
            else:
                self.update({element.tag: element.text})


def xml_to_dict(path: str) -> dict[str, Any]:
    """
    Given a XML file path, returns the contents of the file as a dictionnary
    """
    return XmlDictConfig(ElementTree.parse(path).getroot())



# MIT License

# Copyright (c) 2025 Rodolphe Durand

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import glob
import os
import pathlib
import platform
import shutil
import subprocess

import huggingface_hub


def check_7z():
    """
    Checks if 7zip is installed and if the command is available. If not, raises an error asking the user to install it.
    """
    if shutil.which("7z") is None:
        message = "7zip is needed to unpack the dataset files. Please install it and add command to path."
        if platform.system() == "Linux":
            message += " The following command can be used :\n\n     sudo apt-get install p7zip-full\n"

        raise RuntimeError(message)


def download_benchmark(
    benchmark: str,
    root_dir: pathlib.Path,
    hf_repo_id: str,
    hf_benchmark_path: pathlib.Path,
):
    """
    Download the specified benchmark in the given root directory.

    The dowload uses Hugging Face. The user will be prompted to log in using an acess token
    A Hugging Face account is thus required to download the dataset (even when doing so manually)
    """
    huggingface_hub.login(new_session=False)

    # create root_dir if it does not exist
    root_dir.mkdir(parents=True, exist_ok=True)
    os.chdir(root_dir.absolute())

    # gather all volumes of the split 7z archive (file.7z.001, file.7z.002, ...)
    benchmark_glob = f"{str(hf_benchmark_path / benchmark)}.7z.*"
    huggingface_hub.snapshot_download(
        repo_id=hf_repo_id,
        allow_patterns=benchmark_glob,
        local_dir=root_dir.absolute(),
        repo_type="dataset",
    )

    # downloaded files maintain their original file structure
    # move them back to root_dir and delete empty directories
    for zip_file in (root_dir / hf_benchmark_path).glob("*.*"):
        shutil.move(zip_file, root_dir)
    os.removedirs(hf_benchmark_path)

    # unzip the first archive (7zip automatically joins all volumes)
    subprocess.run(f"7z x {benchmark}.7z.001", shell=True)

    # delete unpacked 7zip files
    for f in glob.glob(f"{benchmark}.7z.*"):
        os.remove(f)



# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
from typing import Union, Callable, Optional

# External imports
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules.transformer import (
    _get_clones,
    _get_seq_len,
    _detect_is_causal_mask,
    TransformerEncoder,
    TransformerDecoder,
)

# Local imports
from .activation import CReLU, MultiheadAttention
from .dropout import Dropout
from .normalization import LayerNorm
from .initialization import complex_xavier_uniform_


class TransformerEncoderLayer(nn.Module):
    r"""TransformerEncoderLayer is made up of self-attn and feedforward network.

    This class is adapted from pytorch :py:class:`torch.nn.TransformerEncoderLayer`

    This standard encoder layer is based on the paper **Attention Is All You Need**.
    Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez,
    Lukasz Kaiser, and Illia Polosukhin. 2017. Attention is all you need. In Advances in
    Neural Information Processing Systems, pages 6000-6010. Users may modify or implement
    in a different way during application.

    If you are implementing a custom layer, you may derive it either from
    the Module or TransformerEncoderLayer class.

    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of the intermediate layer. Default: :py:class:`CReLU`
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        batch_first: If ``True``, then the input and output tensors are provided
            as (batch, seq, feature). Default: ``False`` (seq, batch, feature).
        norm_first: if ``True``, layer norm is done prior to attention and feedforward
            operations, respectively. Otherwise it's done after. Default: ``False`` (after).
        bias: If set to ``False``, ``Linear`` and ``LayerNorm`` layers will not learn an additive
            bias. Default: ``True``.

    Examples:

        .. code-block:: python

            import torchcvnn as c_nn
            import torch

            encoder_layer = c_nn.TransformerEncoderLayer(d_model=512, nhead=8)
            src = torch.rand(10, 32, 512, dtype=torch.complex64)
            out = encoder_layer(src)

        Alternatively, when ``batch_first`` is ``True``:

        .. code-block:: python

            import torchcvnn as c_nn
            import torch

            encoder_layer = c_nn.TransformerEncoderLayer(d_model=512, nhead=8, batch_first=True)
            src = torch.rand(32, 10, 512, dtype=torch.complex64)
            out = encoder_layer(src)

    """

    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: nn.Module = CReLU,
        layer_norm_eps: float = 1e-5,
        batch_first: bool = False,
        norm_first: bool = False,
        bias: bool = True,
        device: torch.device = None,
        dtype: torch.dtype = torch.complex64,
        attn_module=MultiheadAttention,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.self_attn = attn_module(
            d_model,
            nhead,
            dropout=dropout,
            bias=bias,
            batch_first=batch_first,
            **factory_kwargs,
        )
        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward, bias=bias, **factory_kwargs)
        self.dropout = Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model, bias=bias, **factory_kwargs)

        self.norm_first = norm_first
        self.norm1 = LayerNorm(d_model, eps=layer_norm_eps, bias=bias, **factory_kwargs)
        self.norm2 = LayerNorm(d_model, eps=layer_norm_eps, bias=bias, **factory_kwargs)
        self.dropout1 = Dropout(dropout)
        self.dropout2 = Dropout(dropout)

        self.activation = activation()

        self._reset_parameters()

    def _reset_parameters(self):
        complex_xavier_uniform_(self.linear1.weight)
        if self.linear1.bias is not None:
            nn.init.constant_(self.linear1.bias, 0)
        complex_xavier_uniform_(self.linear2.weight)
        if self.linear2.bias is not None:
            nn.init.constant_(self.linear2.bias, 0)

    def __setstate__(self, state):
        super().__setstate__(state)
        if not hasattr(self, "activation"):
            self.activation = CReLU()

    def forward(
        self,
        src: torch.Tensor,
        src_mask: Optional[torch.Tensor] = None,
        src_key_padding_mask: Optional[torch.Tensor] = None,
        is_causal: bool = False,
    ) -> torch.Tensor:

        x = src
        if self.norm_first:
            x = x + self._sa_block(
                self.norm1(x), src_mask, src_key_padding_mask, is_causal
            )
            x = x + self._ff_block(self.norm2(x))
        else:
            x = x + self._sa_block(x, src_mask, src_key_padding_mask, is_causal)
            x = self.norm1(x)
            x = x + self._ff_block(x)
            x = self.norm2(x)

        return x

    def _sa_block(
        self,
        x: torch.Tensor,
        attn_mask: Optional[torch.Tensor],
        key_padding_mask: Optional[torch.Tensor],
        is_causal: bool,
    ) -> torch.Tensor:
        x = self.self_attn(
            x,
            x,
            x,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            need_weights=False,
            is_causal=is_causal,
        )[0]
        x = self.dropout1(x)
        return x

    def _ff_block(self, x: torch.Tensor) -> torch.Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout2(x)


class TransformerDecoderLayer(nn.Module):
    r"""TransformerDecoderLayer is made up of self-attn, multi-head-attn and feedforward network.

    Adapted from Pytorch :py:class:`torch.nn.TransformerDecoderLayer`.

    This standard decoder layer is based on the paper **Attention Is All You Need**.
    Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez,
    Lukasz Kaiser, and Illia Polosukhin. 2017. Attention is all you need. In Advances in
    Neural Information Processing Systems, pages 6000-6010. Users may modify or implement
    in a different way during application.

    Args:
        d_model: the number of expected features in the input (required).
        nhead: the number of heads in the multiheadattention models (required).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of the intermediate layer. Default: :py:class:`CReLU`
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        batch_first: If ``True``, then the input and output tensors are provided
            as (batch, seq, feature). Default: ``False`` (seq, batch, feature).
        norm_first: if ``True``, layer norm is done prior to self attention, multihead
            attention and feedforward operations, respectively. Otherwise it's done after.
            Default: ``False`` (after).
        bias: If set to ``False``, ``Linear`` and ``LayerNorm`` layers will not learn an additive
            bias. Default: ``True``.

    Examples::

        .. code-block:: python

            import torchcvnn as c_nn
            import torch

            decoder_layer = c_nn.TransformerDecoderLayer(d_model=512, nhead=8)
            memory = torch.rand(10, 32, 512, dtype=torch.complex64)
            tgt = torch.rand(20, 32, 512, dtype=torch.complex64)
            out = decoder_layer(tgt, memory)

        Alternatively, when ``batch_first`` is ``True``:

        .. code-block:: python

            import torchcvnn as c_nn
            import torch

            decoder_layer = c_nn.TransformerDecoderLayer(d_model=512, nhead=8, batch_first=True)
            memory = torch.rand(32, 10, 512, dtype=torch.complex64)
            tgt = torch.rand(32, 20, 512, dtype=torch.complex64)
            out = decoder_layer(tgt, memory)
    """

    __constants__ = ["norm_first"]

    # Adapted from Pytorch TransformerDecoderLayer
    # with CReLU instead of ReLU and dtype=torch.complex64
    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: nn.Module = CReLU,
        layer_norm_eps: float = 1e-5,
        batch_first: bool = False,
        norm_first: bool = False,
        bias: bool = True,
        device=None,
        dtype: torch.dtype = torch.complex64,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.self_attn = MultiheadAttention(
            d_model,
            nhead,
            dropout=dropout,
            batch_first=batch_first,
            bias=bias,
            **factory_kwargs,
        )
        self.multihead_attn = MultiheadAttention(
            d_model,
            nhead,
            dropout=dropout,
            batch_first=batch_first,
            bias=bias,
            **factory_kwargs,
        )
        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward, bias=bias, **factory_kwargs)
        self.dropout = Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model, bias=bias, **factory_kwargs)

        self.norm_first = norm_first
        self.norm1 = LayerNorm(d_model, eps=layer_norm_eps, bias=bias, **factory_kwargs)
        self.norm2 = LayerNorm(d_model, eps=layer_norm_eps, bias=bias, **factory_kwargs)
        self.norm3 = LayerNorm(d_model, eps=layer_norm_eps, bias=bias, **factory_kwargs)
        self.dropout1 = Dropout(dropout)
        self.dropout2 = Dropout(dropout)
        self.dropout3 = Dropout(dropout)

        self.activation = activation()

        self._reset_parameters()

    def _reset_parameters(self):
        complex_xavier_uniform_(self.linear1.weight)
        if self.linear1.bias is not None:
            nn.init.constant_(self.linear1.bias, 0)
        complex_xavier_uniform_(self.linear2.weight)
        if self.linear2.bias is not None:
            nn.init.constant_(self.linear2.bias, 0)

    # Adapted from Pytorch TransformerDecoderLayer
    # with CReLU instead of ReLU
    def __setstate__(self, state):
        if "activation" not in state:
            state["activation"] = CReLU()
        super().__setstate__(state)

    # Same from Pytorch TransformerDecoderLayer
    def forward(
        self,
        tgt: torch.Tensor,
        memory: torch.Tensor,
        tgt_mask: Optional[torch.Tensor] = None,
        memory_mask: Optional[torch.Tensor] = None,
        tgt_key_padding_mask: Optional[torch.Tensor] = None,
        memory_key_padding_mask: Optional[torch.Tensor] = None,
        tgt_is_causal: bool = False,
        memory_is_causal: bool = False,
    ) -> torch.Tensor:
        r"""Pass the inputs (and mask) through the decoder layer.

        Args:
            tgt: the sequence to the decoder layer (required).
            memory: the sequence from the last layer of the encoder (required).
            tgt_mask: the mask for the tgt sequence (optional).
            memory_mask: the mask for the memory sequence (optional).
            tgt_key_padding_mask: the mask for the tgt keys per batch (optional).
            memory_key_padding_mask: the mask for the memory keys per batch (optional).
            tgt_is_causal: If specified, applies a causal mask as ``tgt mask``.
                Default: ``False``.
                Warning:
                ``tgt_is_causal`` provides a hint that ``tgt_mask`` is
                the causal mask. Providing incorrect hints can result in
                incorrect execution, including forward and backward
                compatibility.
            memory_is_causal: If specified, applies a causal mask as
                ``memory mask``.
                Default: ``False``.
                Warning:
                ``memory_is_causal`` provides a hint that
                ``memory_mask`` is the causal mask. Providing incorrect
                hints can result in incorrect execution, including
                forward and backward compatibility.

        Shape:
            see the docs in :class:`~torch.nn.Transformer`.
        """
        # see Fig. 1 of https://arxiv.org/pdf/2002.04745v1.pdf

        x = tgt
        if self.norm_first:
            x = x + self._sa_block(
                self.norm1(x), tgt_mask, tgt_key_padding_mask, tgt_is_causal
            )
            x = x + self._mha_block(
                self.norm2(x),
                memory,
                memory_mask,
                memory_key_padding_mask,
                memory_is_causal,
            )
            x = x + self._ff_block(self.norm3(x))
        else:
            x = self.norm1(
                x + self._sa_block(x, tgt_mask, tgt_key_padding_mask, tgt_is_causal)
            )
            x = self.norm2(
                x
                + self._mha_block(
                    x, memory, memory_mask, memory_key_padding_mask, memory_is_causal
                )
            )
            x = self.norm3(x + self._ff_block(x))

        return x

    # self-attention block
    # Same from Pytorch TransformerDecoderLayer
    def _sa_block(
        self,
        x: torch.Tensor,
        attn_mask: Optional[torch.Tensor],
        key_padding_mask: Optional[torch.Tensor],
        is_causal: bool = False,
    ) -> torch.Tensor:
        x = self.self_attn(
            x,
            x,
            x,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            is_causal=is_causal,
            need_weights=False,
        )[0]
        return self.dropout1(x)

    # multihead attention block
    # Same from Pytorch TransformerDecoderLayer
    def _mha_block(
        self,
        x: torch.Tensor,
        mem: torch.Tensor,
        attn_mask: Optional[torch.Tensor],
        key_padding_mask: Optional[torch.Tensor],
        is_causal: bool = False,
    ) -> torch.Tensor:
        x = self.multihead_attn(
            x,
            mem,
            mem,
            attn_mask=attn_mask,
            key_padding_mask=key_padding_mask,
            is_causal=is_causal,
            need_weights=False,
        )[0]
        return self.dropout2(x)

    # feed forward block
    # Same from Pytorch TransformerDecoderLayer
    def _ff_block(self, x: torch.Tensor) -> torch.Tensor:
        x = self.linear2(self.dropout(self.activation(self.linear1(x))))
        return self.dropout3(x)


class Transformer(nn.Module):
    r"""A transformer model.

    Adapted from :py:class:`torch.nn.Transformer`.

    User is able to modify the attributes as needed. The architecture
    is based on the paper **Attention Is All You Need**. Ashish Vaswani, Noam Shazeer,
    Niki Parmar, Jakob Uszkoreit, Llion Jones, Aidan N Gomez, Lukasz Kaiser, and
    Illia Polosukhin. 2017. Attention is all you need. In Advances in Neural Information
    Processing Systems, pages 6000-6010.


    The :py:class:`MultiheadAttention` implementation is based on the paper **Building blocks for a complex-valued
    transformer architecture**. Florian Eilers, Xiaoyi Jiang. 2023. In International Conference on Acoustics, Speech,
    and Signal Processing (ICASSP).


    Args:
        d_model: the number of expected features in the encoder/decoder inputs (default=512).
        nhead: the number of heads in the multiheadattention models (default=8).
        num_encoder_layers: the number of sub-encoder-layers in the encoder (default=6).
        num_decoder_layers: the number of sub-decoder-layers in the decoder (default=6).
        dim_feedforward: the dimension of the feedforward network model (default=2048).
        dropout: the dropout value (default=0.1).
        activation: the activation function of encoder/decoder intermediate layer. Default: :py:class:`CReLU`.
        custom_encoder: custom encoder (default=None).
        custom_decoder: custom decoder (default=None).
        layer_norm_eps: the eps value in layer normalization components (default=1e-5).
        batch_first: If ``True``, then the input and output tensors are provided
            as (batch, seq, feature). Default: ``False`` (seq, batch, feature).
        norm_first: if ``True``, encoder and decoder layers will perform LayerNorms before
            other attention and feedforward operations, otherwise after. Default: ``False`` (after).
        bias: If set to ``False``, ``Linear`` and ``LayerNorm`` layers will not learn an additive
            bias. Default: ``True``.

    Examples:

        .. code-block:: python

            import torchcvnn as c_nn
            import torch

            transformer_model = c_nn.Transformer(nhead=16, num_encoder_layers=12)
            src = torch.rand((10, 32, 512), dtype=torch.complex64)
            tgt = torch.rand((20, 32, 512), dtype=torch.complex64)
            out = transformer_model(src, tgt)

    """

    def __init__(
        self,
        d_model: int = 512,
        nhead: int = 8,
        num_encoder_layers: int = 6,
        num_decoder_layers: int = 6,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: nn.Module = CReLU,
        layer_norm_eps: float = 1e-5,
        batch_first: bool = False,
        norm_first: bool = False,
        bias: bool = True,
        device=None,
        dtype: torch.dtype = torch.complex64,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()

        encoder_layer = TransformerEncoderLayer(
            d_model,
            nhead,
            dim_feedforward,
            dropout,
            activation,
            layer_norm_eps,
            batch_first,
            norm_first,
            bias,
            **factory_kwargs,
        )
        encoder_norm = LayerNorm(
            d_model, eps=layer_norm_eps, bias=bias, **factory_kwargs
        )
        self.encoder = TransformerEncoder(
            encoder_layer, num_encoder_layers, encoder_norm
        )

        decoder_layer = TransformerDecoderLayer(
            d_model,
            nhead,
            dim_feedforward,
            dropout,
            activation,
            layer_norm_eps,
            batch_first,
            norm_first,
            bias,
            **factory_kwargs,
        )
        decoder_norm = LayerNorm(
            d_model, eps=layer_norm_eps, bias=bias, **factory_kwargs
        )
        self.decoder = TransformerDecoder(
            decoder_layer, num_decoder_layers, decoder_norm
        )

        self.d_model = d_model
        self.nhead = nhead

        self.batch_first = batch_first

    def forward(
        self,
        src: torch.Tensor,
        tgt: torch.Tensor,
        src_mask: Optional[torch.Tensor] = None,
        tgt_mask: Optional[torch.Tensor] = None,
        memory_mask: Optional[torch.Tensor] = None,
        src_key_padding_mask: Optional[torch.Tensor] = None,
        tgt_key_padding_mask: Optional[torch.Tensor] = None,
        memory_key_padding_mask: Optional[torch.Tensor] = None,
        src_is_causal: Optional[bool] = None,
        tgt_is_causal: Optional[bool] = None,
        memory_is_causal: bool = False,
    ) -> torch.Tensor:
        memory = self.encoder(
            src,
            mask=src_mask,
            src_key_padding_mask=src_key_padding_mask,
            is_causal=src_is_causal,
        )
        output = self.decoder(
            tgt,
            memory,
            tgt_mask=tgt_mask,
            memory_mask=memory_mask,
            tgt_key_padding_mask=tgt_key_padding_mask,
            memory_key_padding_mask=memory_key_padding_mask,
            tgt_is_causal=tgt_is_causal,
            memory_is_causal=memory_is_causal,
        )
        return output



# MIT License

# Copyright (c) 2024 Chengfang Ren, Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
import pathlib
from typing import Tuple, Any

# External imports
from torch.utils.data import Dataset
import numpy as np


class Bretigny(Dataset):
    r"""
    Bretigny Dataset

    Arguments:
        root: the root directory containing the npz files for Bretigny
        fold: train (70%), valid (15%), or test (15%)
        transform : the transform applied the cropped image
        balanced: whether or not to use balanced labels
        patch_size: the dimensions of the patches to consider (rows, cols)
        patch_stride: the shift between two consecutive patches, default:patch_size

    Note:
        An example usage :

        .. code-block:: python

            import torchcvnn
            from torchcvnn.datasets import Bretigny

            dataset = Bretigny(
                rootdir, fold="train", patch_size=((128, 128)), transform=lambda x: np.abs(x)
            )
            X, y = dataset[0]

        Displayed below are the train, valid and test parts with the labels overlayed

        .. image:: ../assets/datasets/bretigny_train.png
           :alt: Train fold
           :width: 60%
        .. image:: ../assets/datasets/bretigny_valid.png
           :alt: Valid fold
           :width: 15%
        .. image:: ../assets/datasets/bretigny_test.png
           :alt: Test fold
           :width: 15%

    """

    """
    Class names
    """
    classes = ["0 - Unlabeld", "1 - Forest", "2 - Track", "3 - Urban", "4 - Fields"]

    def __init__(
        self,
        root: str,
        fold: str,
        transform=None,
        balanced: bool = False,
        patch_size: tuple = (128, 128),
        patch_stride: tuple = None,
    ):
        self.root = pathlib.Path(root)
        self.transform = transform

        self.patch_size = patch_size
        self.patch_stride = patch_stride
        if patch_stride is None:
            self.patch_stride = patch_size

        # Preload the data
        sar_filename = self.root / "bretigny_seg.npz"
        if not sar_filename.exists():
            raise RuntimeError(f"Cannot find the file {sar_filename}")
        sar_data = np.load(sar_filename)
        self.HH, self.HV, self.VV = sar_data["HH"], sar_data["HV"], sar_data["VV"]

        if balanced:
            label_filename = self.root / "bretigny_seg_4ROI_balanced.npz"
        else:
            label_filename = self.root / "bretigny_seg_4ROI.npz"
        if not label_filename.exists():
            raise RuntimeError(f"Cannot find the label file {label_filename}")
        self.labels = np.load(label_filename)["arr_0"]

        if not fold in ["train", "valid", "test"]:
            raise ValueError(
                f"Unrecognized fold {fold}. Should be either train, valid or test"
            )

        # Crop the data with respect to the fold
        if fold == "train":
            col_start = 0
            col_end = int(0.70 * self.HH.shape[1])
        elif fold == "valid":
            col_start = int(0.70 * self.HH.shape[1]) + 1
            col_end = int(0.85 * self.HH.shape[1])
        else:
            col_start = int(0.85 * self.HH.shape[1]) + 1
            col_end = self.HH.shape[1]

        self.HH = self.HH[:, col_start:col_end]
        self.HV = self.HV[:, col_start:col_end]
        self.VV = self.VV[:, col_start:col_end]
        self.labels = self.labels[:, col_start:col_end]

        # Precompute the dimension of the grid of patches
        nrows = self.HH.shape[0]
        ncols = self.HH.shape[1]

        nrows_patch, ncols_patch = self.patch_size
        row_stride, col_stride = self.patch_stride

        self.nsamples_per_rows = (nrows - nrows_patch) // row_stride + 1
        self.nsamples_per_cols = (ncols - ncols_patch) // col_stride + 1

    def __len__(self) -> int:
        """
        Returns the total number of patches in the whole image.

        Returns:
            the total number of patches in the dataset
        """
        return self.nsamples_per_rows * self.nsamples_per_cols

    def __getitem__(self, idx) -> Tuple[Any, Any]:
        """
        Returns the indexes patch.

        Arguments:
            idx (int): Index

        Returns:
            tuple: (patch, labels) where patch contains the 3 complex valued polarization HH, HV, VV and labels contains the aligned semantic labels
        """
        row_stride, col_stride = self.patch_stride
        start_row = (idx // self.nsamples_per_cols) * row_stride
        start_col = (idx % self.nsamples_per_cols) * col_stride
        num_rows, num_cols = self.patch_size
        patches = [
            patch[
                start_row : (start_row + num_rows), start_col : (start_col + num_cols)
            ]
            for patch in [self.HH, self.HV, self.VV]
        ]
        patches = np.stack(patches)
        if self.transform is not None:
            patches = self.transform(patches)

        labels = self.labels[
            start_row : (start_row + num_rows), start_col : (start_col + num_cols)
        ]

        return patches, labels



# MIT License

# Copyright (c) 2024 Quentin Gabot

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
import pathlib
from typing import Tuple, Any

# External imports
import numpy as np
from torch.utils.data import Dataset
from PIL import Image

# Local imports
from .alos2 import ALOSDataset


class PolSFDataset(Dataset):
    r"""
    The Polarimetric SAR dataset with the labels provided by
    https://ietr-lab.univ-rennes1.fr/polsarpro-bio/san-francisco/

    We expect the data to be already downloaded and available on your drive.

    Arguments:
        root: the top root dir where the data are expected
        transform : the transform applied the cropped image
        patch_size: the dimensions of the patches to consider (rows, cols)
        patch_stride: the shift between two consecutive patches, default:patch_size

    Note:
        An example usage :

        .. code-block:: python

            import torchcvnn
            from torchcvnn.datasets import PolSFDataset

            def transform_patches(patches):
                # We keep all the patches and get the spectrum
                # from it
                # If you wish, you could filter out some polarizations
                # PolSF provides the four HH, HV, VH, VV
                patches = [np.abs(patchi) for _, patchi in patches.items()]
                return np.stack(patches)

            dataset = PolSFDataset(rootdir, patch_size=((512, 512)), transform=transform_patches
            X, y = dataset[0]

        Displayed below are example patches with patch sizes :math:`512 \times 512`
        with the labels overlayed

        .. figure:: ../assets/datasets/polsf.png
           :alt: Patches from the PolSF dataset
           :width: 100%
           :align: center

    """

    """
    Class names
    """
    classes = [
        "0 - unlabel",
        "1 - Montain",
        "2 - Water",
        "3 - Vegetation",
        "4 - High-Density Urban",
        "5 - Low-Density Urban",
        "6 - Developd",
    ]

    def __init__(
        self,
        root: str,
        transform=None,
        patch_size: tuple = (128, 128),
        patch_stride: tuple = None,
    ):
        self.root = root

        # alos2_url = "https://ietr-lab.univ-rennes1.fr/polsarpro-bio/san-francisco/dataset/SAN_FRANCISCO_ALOS2.zip"
        # labels_url = "https://raw.githubusercontent.com/liuxuvip/PolSF/master/SF-ALOS2/SF-ALOS2-label2d.png"

        crop_coordinates = ((2832, 736), (7888, 3520))
        root = pathlib.Path(root) / "VOL-ALOS2044980750-150324-HBQR1.1__A"
        self.alos_dataset = ALOSDataset(
            root, transform, crop_coordinates, patch_size, patch_stride
        )
        if isinstance(root, str):
            root = pathlib.Path(root)
        self.labels = np.array(Image.open(root.parent / "SF-ALOS2-label2d.png"))[
            ::-1, :
        ].copy()  # copy necessary as otherwise torch.from_numpy does not support
        # negative stride

    def __len__(self) -> int:
        """
        Returns the total number of patches in the while image.

        Returns:
            the total number of patches in the dataset
        """
        return len(self.alos_dataset)

    def __getitem__(self, idx) -> Tuple[Any, Any]:
        """
        Returns the indexes patch.

        Arguments:
            idx (int): Index

        Returns:
            tuple: (patch, labels) where patch contains the 4 complex valued polarization HH, HV, VH, VV and labels contains the aligned semantic labels
        """
        alos_patch = self.alos_dataset[idx]

        row_stride, col_stride = self.alos_dataset.patch_stride
        start_row = (idx // self.alos_dataset.nsamples_per_cols) * row_stride

        start_col = (idx % self.alos_dataset.nsamples_per_cols) * col_stride
        num_rows, num_cols = self.alos_dataset.patch_size
        labels = self.labels[
            start_row : (start_row + num_rows), start_col : (start_col + num_cols)
        ]

        return alos_patch, labels



# MIT License

# Copyright (c) 2024 Quentin Gabot

# The following is an adaptation of the initialization code from
# PyTorch library for the complex-valued neural networks

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# External imports
import torch
import torch.nn as nn


class ComplexMSELoss(nn.modules.loss._Loss):
    r"""A custom implementation of Mean Squared Error (MSE) Loss for complex-valued inputs.

    This loss function is designed to compute the mean squared error between two sets of complex-valued tensors. Unlike the standard MSE loss which directly computes the square of the difference between predicted values and true values for real numbers, this implementation adapts to complex numbers by calculating the squared magnitude of the difference between the complex predicted and true values before averaging.

    For complex-valued numbers, the MSE is defined as:
        MSE = mean(|y_true - y_pred|^2)
    where |y_true - y_pred| denotes the magnitude of the complex difference between the true and predicted values. This formula ensures that both the real and imaginary parts of the complex numbers are considered in the loss calculation.

    For real-valued numbers, the standard MSE formula is:
        MSE = mean((y_true - y_pred)^2)
    where (y_true - y_pred)^2 is the square of the difference between the true and predicted values.

    Parameters:
        size_average (bool, optional): Deprecated (use reduction). By default, the losses are averaged over each loss element in the batch. Note: when (reduce is False), size_average is ignored. Default is None.
        reduce (bool, optional): Deprecated (use reduction). By default, the losses are averaged or summed over observations for each minibatch depending on reduction. Note: when (reduce is False), size_average is ignored. Default is None.
        reduction (str, optional): Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'. 'none': no reduction will be applied, 'mean': the sum of the output will be divided by the number of elements in the output, 'sum': the output will be summed. Default is 'mean'.

    Inputs:
        y_pred (torch.Tensor): The predicted values, can be a complex tensor.
        y_true (torch.Tensor): The ground truth values, must have the same shape as the predicted values, can be a complex tensor.

    Returns:
        torch.Tensor: The calculated mean squared error loss.

    Examples:
        >>> loss = ComplexMSELoss()
        >>> y_pred = torch.tensor([1+1j, 2+2j], dtype=torch.complex64)
        >>> y_true = torch.tensor([1+2j, 2+3j], dtype=torch.complex64)
        >>> output = loss(y_pred, y_true)
        >>> print(output)
    """

    def __init__(self, size_average=None, reduce=None, reduction: str = "mean") -> None:
        super().__init__(size_average, reduce, reduction)

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        # Calculate Mean Square Loss for complex numbers
        return torch.mean(torch.square(torch.abs(y_true - y_pred)))



# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class AnnFile:
    r"""

    From the documentation :

    The annotation file (.ann) is a keyword/value ASCII file in which the value on the right of the equals sign
    corresponds to the keyword on the left of the equals sign. The number of keywords may change with time, so the line
    number should not be assumed to be constant for any given keyword.

    In addition, the spacing between keywords and
    values may change. The units are given in parenthesis between the keyword and equal sign, and may change from
    annotation file to annotation file and within each annotation file.

    Comments are indicated by semicolons (;), and
    may occur at the beginning of a line, or at the middle of a line (everything after the semicolon on that line is a
    comment). The length of each text line is variable, and ends with a carriage return. There may be lines with
    just a carriage return or spaces and a carriage return.


    """

    def __init__(self, filename):
        self.filename = filename
        self.parameters = []
        self.data = self.read()

    def read(self):
        with open(self.filename, "r") as f:
            for l in f:
                l = l.strip()
                if l.startswith(";"):
                    # Skip comments
                    continue
                if len(l) != 0:
                    # Skip empty lines
                    fields = l.split(" = ")
                    if len(fields) != 2:
                        continue
                    key = fields[0].strip()
                    # key might be :
                    #     ISLR Noise Calibration Term LRTI80                       (-)
                    #     slc_1_1x1_mag.set_cols                                   (pixels)
                    # ....
                    # We split the key by spaces, take all but the last token (the units) and join them with an underscore
                    # to get a key that is easier to use
                    key = "_".join(
                        list(map(lambda s: s.strip().lower(), key.split()))[:-1]
                    ).replace(".", "_")

                    # Now we process the value, which may end with a comment
                    value = fields[1].split(";")[0].strip()
                    try:
                        value = int(value)
                    except ValueError:
                        try:
                            value = float(value)
                        except ValueError:
                            pass
                    setattr(self, key, value)
                    self.parameters.append(key)

    def __repr__(self):
        myrep = f"AnnFile({self.filename})\n"
        for p in self.parameters:
            myrep += f"     {p} = {getattr(self, p)}\n"
        return myrep



# MIT License

# Copyright (c) 2023 Jérémie Levi, Victor Dhédin, Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# External imports
import torch

# Local imports
import torchcvnn.nn as c_nn


def test_mpool2d():
    mpool = c_nn.MaxPool2d(kernel_size=2, stride=2, padding=0)

    N = 10
    inputs = torch.tensor([[i + j * 1j for i in range(N)] for j in range(N)])
    inputs = inputs.unsqueeze(0).unsqueeze(0)
    outputs = mpool(inputs)

    expected_outputs = (
        torch.tensor(
            [
                [2 * i + 1 + (2 * j + 1) * 1j for i in range(N // 2)]
                for j in range(N // 2)
            ]
        )
        .unsqueeze(0)
        .unsqueeze(0)
    )
    assert torch.allclose(outputs, expected_outputs)


def test_avgpool2d():
    avgpool = c_nn.AvgPool2d(kernel_size=3, stride=1, padding=1)

    N = 10
    inputs = torch.tensor([[i + j * 1j for i in range(N)] for j in range(N)])
    inputs = inputs.unsqueeze(0).unsqueeze(0)
    outputs = avgpool(inputs)

    # With kernel_size=3, stride=1, padding=1, output is the same shape
    # as the input
    assert inputs.shape == outputs.shape

    def num_series(row_or_col):
        return 3 if row_or_col > 0 and row_or_col < N - 1 else 2

    expected_outputs = (
        torch.tensor(
            [
                [
                    (num_series(j) * (i - 1 + i + i + 1)) / 9.0 * 1j
                    + (num_series(i) * (j - 1 + j + j + 1)) / 9.0
                    for j in range(N)
                ]
                for i in range(N)
            ]
        )
        .unsqueeze(0)
        .unsqueeze(0)
    )
    # Handle the border cases
    expected_outputs[0, 0, 1:-1, 0] += 3.0 / 9.0
    expected_outputs[0, 0, 0, 0] += 2.0 / 9.0
    expected_outputs[0, 0, -1, 0] += 2.0 / 9.0
    expected_outputs[0, 0, 1:-1, -1] -= 3.0 * N / 9.0
    expected_outputs[0, 0, 0, -1] -= 2.0 * N / 9.0
    expected_outputs[0, 0, -1, -1] -= 2.0 * N / 9.0

    expected_outputs[0, 0, 0, 1:-1] += 3.0 / 9.0 * 1j
    expected_outputs[0, 0, 0, 0] += 2.0 / 9.0 * 1j
    expected_outputs[0, 0, 0, -1] += 2.0 / 9.0 * 1j
    expected_outputs[0, 0, -1, 1:-1] -= 3.0 * N / 9.0 * 1j
    expected_outputs[0, 0, -1, 0] -= 2.0 * N / 9.0 * 1j
    expected_outputs[0, 0, -1, -1] -= 2.0 * N / 9.0 * 1j

    assert torch.allclose(outputs, expected_outputs)


if __name__ == "__main__":
    test_mpool2d()
    test_avgpool2d()



# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
from typing import Union
import pathlib

# Local imports
from . import parse_utils

descriptor_format = [
    ("record_sequence_number", 0, 4, "B", 1),
    ("first_record_subtype_code", 4, 1, "B", 192),
    ("record_type_code", 5, 1, "B", 192),
    ("second_subtype_code", 6, 1, "B", 18),
    ("third_subtype_code", 7, 1, "B", 18),
    ("length_record", 8, 4, "B", 360),
    ("flag", 12, 2, "A", "A "),
    # ("blanks", 14, 2, "A", "  "),
    ("superstructure_doc_id", 16, 12, "A", "CEOS-SAR    "),
    ("superstructure_doc_rev_level", 28, 2, "A", " A"),
    ("superstructure_fmt_rev_level", 30, 2, "A", " A"),
    ("software_release_level", 32, 12, "A", None),
    ("physical_volume_id", 44, 16, "A", None),
    ("logical_volume_id", 60, 16, "A", None),
    ("volume_set_id", 76, 16, "A", None),
    ("total_number_volumes", 92, 2, "I", 1),
    ("physical_volume_seq_num_first", 94, 2, "I", 1),
    ("physical_volume_seq_num_last", 96, 2, "I", 1),
    ("physical_volume_seq_num_cur", 98, 2, "I", 1),
    ("file_number", 100, 4, "I", None),
    ("logical_volume_within_volume", 104, 4, "I", None),
    ("logical_volume_within_physical", 108, 4, "I", None),
    ("logical_volume_creation_date", 112, 8, "A", None),  # YYYYMMDD
    ("logical_volume_creation_time", 120, 8, "A", None),  # HHMMSSXX
    ("logical_volume_creation_country", 128, 12, "A", None),
    ("logical_volume_creation_agency", 140, 8, "A", None),
    ("logical_volume_generation_facility", 148, 12, "A", None),
    ("number_of_file_pointer_records", 160, 4, "I", None),
    ("number_of_text_records", 164, 4, "I", None),
    # Volume descriptor spare A92 + local use segment A100
]
volume_descriptor_record_length = 360

file_pointer_format = [
    ("record_number", 0, 4, "B", None),
    ("record_type", 5, 1, "B", None),
    ("reference_id", 20, 16, "A", None),
    ("reference_file_class_code", 64, 4, "A", None),
    # ("number_of_records", 100, 8, "I", None),
    # ("number_of_physical", 140, 2, "I", None),
]
file_pointer_record_length = 360

text_records_format = [
    ("record_number", 0, 4, "B", None),
    ("product_id", 16, 40, "A", None),
    ("scene_id", 156, 40, "A", None),
    ("scene_location_id", 196, 40, "A", None),
]
text_record_length = 360


class VolFile:
    r"""
    Processing a Volume Directory file in the CEOS format. The parsed
    informations can be accessed through the attributes `descriptor_records`,
    `file_pointer_records` and `text_records`

    Arguments:
        filepath: the path to the volume directory file
    """

    def __init__(self, filepath: Union[str, pathlib.Path]):
        self.descriptor_records = {}
        self.file_pointer_records = []
        self.text_record = {}

        with open(filepath, "rb") as fh:
            # Parsing the volume descriptor
            fh_offset = 0
            fh_offset = parse_utils.parse_from_format(
                fh,
                self.descriptor_records,
                descriptor_format,
                1,
                volume_descriptor_record_length,
                fh_offset,
            )

            # Parsing the file pointer
            number_of_file_pointer_records = self.descriptor_records[
                "number_of_file_pointer_records"
            ]
            fh_offset = parse_utils.parse_from_format(
                fh,
                self.file_pointer_records,
                file_pointer_format,
                number_of_file_pointer_records,
                file_pointer_record_length,
                fh_offset,
            )

            # Parsing the file pointer
            fh_offset = parse_utils.parse_from_format(
                fh,
                self.text_record,
                text_records_format,
                1,
                text_record_length,
                fh_offset,
            )

    @property
    def num_polarizations(self):
        return len(
            [
                True
                for fp in self.file_pointer_records
                if fp["reference_file_class_code"] == "IMOP"
            ]
        )

    def __repr__(self):
        descriptor_txt = parse_utils.format_dictionary(self.descriptor_records, 1)
        # text_txt = parse_utils.format_dictionary(self.text_records, 1)
        fp_texts = ""
        for i, fi in enumerate(self.file_pointer_records):
            fp_texts += f"File pointer {i} : \n" + parse_utils.format_dictionary(fi, 2)
            fp_texts += "\n"

        text_texts = parse_utils.format_dictionary(self.text_record, 1)
        return f"""
Descriptor:
{descriptor_txt}
File pointers : {len(self.file_pointer_records)} records
{fp_texts}

Text records:
{text_texts}
        """



# MIT License

# Copyright (c) 2025 Huy Nguyen

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard imports
from typing import Tuple, Dict
from types import ModuleType

# External imports
import torch
import numpy as np


def polsar_dict_to_array(x: np.ndarray | torch.Tensor | Dict[str, np.ndarray]) -> np.ndarray | torch.Tensor:
    """
    Convert a dictionary of numpy arrays to a stacked array.

    Args:
        x (np.ndarray | torch.Tensor | Dict[str, np.ndarray]): The input data. 
        It can be a single numpy array or PyTorch tensor, or a dictionary where keys are
        one of 'HH', 'HV', 'VH', 'VV' and values are arrays.

    Returns:
        np.ndarray | torch.Tensor: A stacked array from the dictionary's values if input is a dictionary,
        otherwise returns the input unchanged.
    
    Raises:
        AssertionError: If any key in the dictionary is not one of 'HH', 'HV', 'VH', 'VV'.
    """
    if isinstance(x, Dict):
        for k,v in x.items():
            assert k in ['HH', 'HV', 'VH', 'VV'], f'Invalid key {k} in input'
            assert isinstance(v, np.ndarray), "Values must be numpy arrays"
        return np.stack([
            v for v in x.values()
        ])
    return x


def check_input(x: np.ndarray | torch.Tensor) -> np.ndarray | torch.Tensor:
    """Ensure image is in CHW format for 2D-tensors, convert if necessary.
    
    Args:
        x (np.ndarray or torch.Tensor): Input image to check/convert format
        
    Returns:
        np.ndarray or torch.Tensor: Image in CHW format
        
    Raises:
        TypeError: If input is not numpy array or torch tensor
        
    Example:
        >>> img = np.zeros((64, 64))  # HWC format
        >>> chw_img = check_input(img)  # Converts to (1, 64, 64)
    """
    if not isinstance(x, (np.ndarray, torch.Tensor)):
        raise TypeError("Element should be a numpy array or a tensor")
    if len(x.shape) == 2:
        return x[np.newaxis, :, :]
    return x


def log_normalize_amplitude(
    x: np.ndarray | torch.Tensor, 
    backend: ModuleType, 
    keep_phase: bool,
    min_value: float,
    max_value: float,
) -> np.ndarray | torch.Tensor:
    """
    Normalize the amplitude of a complex signal with logarithmic scaling.
    
    Args:
        x: Input array or tensor containing complex numbers. The type can be either numpy ndarray or PyTorch Tensor.
        backend: Module providing mathematical functions, allowing compatibility with numpy or PyTorch.
        keep_phase: Boolean indicating whether to retain the original phase of the input signal.
        max_value: Maximum amplitude value for normalization.
        min_value: Minimum amplitude value for normalization.

    Returns:
        A numpy ndarray or torch.Tensor containing the log-normalized amplitude, optionally with the original phase.
    """
    assert backend.__name__ in ["numpy", "torch"], "Backend must be numpy or torch"
    amplitude = backend.abs(x)
    phase = backend.angle(x)
    amplitude = backend.clip(amplitude, min_value, max_value)
    transformed_amplitude = (
        backend.log10(amplitude / min_value)
    ) / (np.log10(max_value / min_value))
    if keep_phase:
        return transformed_amplitude * backend.exp(1j * phase)
    else:
        return transformed_amplitude


def applyfft2_np(x: np.ndarray, axis: Tuple[int, ...]) -> np.ndarray:
    """Apply 2D Fast Fourier Transform to image.
    
    Args:
        x (np.ndarray): Input array to apply FFT to
        axis (Tuple[int, ...]): Axes over which to compute the FFT
        
    Returns:
        np.ndarray: The Fourier transformed array
    """
    return np.fft.fftshift(np.fft.fft2(x, axes=axis), axes=axis)


def applyifft2_np(x: np.ndarray, axis: Tuple[int, ...]) -> np.ndarray:
    """Apply 2D inverse Fast Fourier Transform to image.
    
    Args:
        x (np.ndarray): Input array to apply IFFT to
        axis (Tuple[int, ...]): Axes over which to compute the IFFT
        
    Returns:
        np.ndarray: The inverse Fourier transformed array
    """
    return np.fft.ifft2(np.fft.ifftshift(x, axes=axis), axes=axis)


def applyfft2_torch(x: torch.Tensor, dim: Tuple[int, ...] = (-2, -1)) -> torch.Tensor:
    """Apply 2D Fast Fourier Transform to image.
    
    Args:
        x (np.ndarray): Input array to apply FFT to
        dim (Tuple[int, ...]): Dimensions over which to compute the FFT. Default is (-2, -1).
        
    Returns:
        torch.Tensor: The Fourier transformed array
    """
    return torch.fft.fftshift(torch.fft.fft2(x, dim=dim), dim=dim)


def applyifft2_torch(x: torch.Tensor, dim: Tuple[int, ...] = (-2, -1)) -> torch.Tensor:
    """Apply 2D inverse Fast Fourier Transform to image.
    
    Args:
        x (torch.Tensor): Input tensor to apply IFFT to
        dim (Tuple[int, ...]): Dimensions over which to compute the IFFT. Default is (-2, -1).
        
    Returns:
        torch.Tensor: The inverse Fourier transformed array
    """
    return torch.fft.ifft2(torch.fft.ifftshift(x, dim=dim), dim=dim)


def get_padding(current_size: int, target_size: int) -> Tuple[int, ...]:
    """Calculate padding required to reach target size from current size.
    
    Calculates padding values for both sides of an axis to reach a target size.
    Handles both even and odd target sizes by adjusting padding distribution.
    
    Args:
        current_size (int): Current dimension size
        target_size (int): Desired dimension size after padding
        
    Returns:
        Tuple[int, ...]: Padding values for (before, after) positions
        
    Example:
        >>> get_padding(5, 7)  # Pad 5->7
        (1, 1)  # Add 1 padding on each side
        >>> get_padding(3, 6)  # Pad 3->6 (even target)
        (2, 1)  # More padding before for even targets
    
    Note:
        For even target sizes, the padding is distributed with one extra
        pad value before the content to maintain proper centering.
    """
    # Adjust offset for even-sized targets or odd-sized targets
    offset = 1 if target_size % 2 == 0 else 0
    # Calculate total padding needed
    pad_total = target_size - current_size
    # Calculate padding before, accounting for even-size offset
    pad_before = (pad_total + offset) // 2
    # Calculate padding after as remainder
    pad_after = pad_total - pad_before
    return pad_before, pad_after


def padifneeded(
    x: np.ndarray | torch.Tensor, 
    min_height: int, 
    min_width: int, 
    border_mode: str = "constant",
    pad_value: float = 0
) -> np.ndarray | torch.Tensor:
    """Pad image if smaller than desired size.

    This function pads an image with zeros if its dimensions are smaller than the specified
    minimum height and width. The padding is added equally on both sides where possible.

    Args:
        x (Union[np.ndarray, torch.Tensor]): Input image tensor/array with shape (C,H,W)
        min_height (int): Minimum required height after padding
        min_width (int): Minimum required width after padding 
        border_mode (str): Padding mode ('constant', 'reflect', 'replicate', etc.)
        pad_value (float): Value used for padding when border_mode is 'constant'. Default: 0

    Returns:
        Union[np.ndarray, torch.Tensor]: Padded image if dimensions were smaller than 
        minimum required, otherwise returns original image unchanged

    Example:
        >>> img = torch.randn(3, 50, 60)  # RGB image 50x60
        >>> padded = padifneeded(img, 64, 64, 'constant')  # Pads to 64x64
        >>> padded.shape
        torch.Size([3, 64, 64])
    """
    _, h, w = x.shape
    # Calculate padding sizes
    top_pad, bottom_pad = get_padding(h, min_height)
    left_pad, right_pad = get_padding(w, min_width)
    padding = [
        top_pad, # top
        bottom_pad,  # bottom
        left_pad,  # left
        right_pad,  # right
    ]
    # Return original if no padding needed
    if all(p <= 0 for p in padding):
        return x

    padding = [max(0, p) for p in padding]
    if isinstance(x, np.ndarray):
        return np.pad(
            x,
            ((0, 0), (padding[0], padding[1]), (padding[2], padding[3])),
            mode=border_mode,
            constant_values=pad_value
        )
    return torch.nn.functional.pad(
        x,
        (padding[2], padding[3], padding[0], padding[1]),
        mode=border_mode,
        value=pad_value
    )


def center_crop(x: np.ndarray | torch.Tensor, height: int, width: int) -> np.ndarray | torch.Tensor:
    """
    Center crops an image to the specified dimensions.

    This function takes an image and crops it to the specified height and width, 
    centered around the middle of the image. If the requested dimensions are larger 
    than the image, it will use the maximum possible size.

    Args:
        x (Union[np.ndarray, torch.Tensor]): Input image tensor/array with shape (C, H, W)
        height (int): Desired height of the cropped image
        width (int): Desired width of the cropped image

    Returns:
        Union[np.ndarray, torch.Tensor]: Center cropped image with shape (C, height, width)

    Example:
        >>> img = torch.randn(3, 100, 100)  # RGB image 100x100
        >>> cropped = center_crop(img, 60, 60)  # Returns center 60x60 crop
        >>> cropped.shape
        torch.Size([3, 60, 60])
    """
    l_h = max(0, x.shape[1] // 2 - height // 2)
    l_w = max(0, x.shape[2] // 2 - width // 2)
    r_h = l_h + height
    r_w = l_w + width
    return x[:, l_h:r_h, l_w:r_w]


def rescale_intensity(image: np.ndarray, in_range: Tuple[float, float], out_range: Tuple[float, float]) -> np.ndarray:
    """Rescale the intensity of an image to a new range.
    
    Args:
        image (np.ndarray): Input image
        in_range (Tuple[float, float]): Input range (min, max) to scale from
        out_range (Tuple[float, float]): Output range (min, max) to scale to
        
    Returns:
        np.ndarray: Rescaled image
    """
    imin, imax = in_range
    omin, omax = out_range
    # Clip the input image to the input range
    image = np.clip(image, imin, imax)
    # Scale to [0, 1]
    if imin != imax:
        image = (image - imin) / (imax - imin)
        # Scale to output range
        return image * (omax - omin) + omin
    else:
        return np.clip(image, omin, omax)


def equalize(image: np.ndarray, plower: int = None, pupper: int = None) -> np.ndarray:
    """Automatically adjust contrast of the SAR image

    Args:
        image (np.ndarray): Image in complex
        plower (int, optional): lower percentile. Defaults to None.
        pupper (int, optional): upper percentile. Defaults to None.

    Returns:
        np.ndarray: Image equalized
    """
    image = np.log10(np.abs(image) + np.spacing(1))
    if not plower:
        vlower, vupper = np.percentile(image, (2, 98))
    else:
        vlower, vupper = np.percentile(image, (plower, pupper))
        
    return np.round(rescale_intensity(image, in_range=(vlower, vupper), out_range=(0, 1)) * 255).astype(np.uint8)



# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# External imports
import torch
from torch.nn.modules.transformer import TransformerDecoder, TransformerEncoder

# Local imports
import torchcvnn.nn as c_nn


def test_multihead_scaleddotproduct_selfattention():
    nhead = 8
    seq_len = 10
    batch_size = 32
    num_features = 512

    multihead_attn = c_nn.MultiheadAttention(embed_dim=num_features, num_heads=nhead)
    src = torch.rand(seq_len, batch_size, num_features, dtype=torch.complex64)
    attn_output, attn_output_weights = multihead_attn(src, src, src)

    assert attn_output.shape == (seq_len, batch_size, num_features)


# def test_multihead_masks():
#    nheads = 8
#
#    src_seq_len = 10
#    tgt_seq_len = 20
#
#    embed_dim = 16  # multiple of nheads, so that head_dim = embed_dim // nheads
#    kdim = 12
#    vdim = 13
#
#    batch_size = 32
#
#    # Batch_first = False
#    query = torch.rand(tgt_seq_len, batch_size, embed_dim, dtype=torch.complex64)
#    key = torch.rand(src_seq_len, batch_size, kdim, dtype=torch.complex64)
#    value = torch.rand(src_seq_len, batch_size, vdim, dtype=torch.complex64)
#
#    key_padding_mask = torch.rand(batch_size, src_seq_len) > 0.5
#
#    multihead_attn = c_nn.MultiheadAttention(
#        embed_dim=embed_dim, num_heads=nheads, kdim=kdim, vdim=vdim
#    )
#    attn_output, attn_output_weights = multihead_attn(
#        query, key, value, key_padding_mask=key_padding_mask
#    )
#
#    assert attn_output.shape == (tgt_seq_len, batch_size, embed_dim)


def test_multihead_scaleddotproduct():
    nheads = 8

    src_seq_len = 10
    tgt_seq_len = 20

    embed_dim = 16  # multiple of nheads, so that head_dim = embed_dim // nheads
    kdim = 12
    vdim = 13

    batch_size = 32

    # Batch_first = False
    query = torch.rand(tgt_seq_len, batch_size, embed_dim, dtype=torch.complex64)
    key = torch.rand(src_seq_len, batch_size, kdim, dtype=torch.complex64)
    value = torch.rand(src_seq_len, batch_size, vdim, dtype=torch.complex64)

    multihead_attn = c_nn.MultiheadAttention(
        embed_dim=embed_dim, num_heads=nheads, kdim=kdim, vdim=vdim
    )
    attn_output, attn_output_weights = multihead_attn(query, key, value)

    assert attn_output.shape == (tgt_seq_len, batch_size, embed_dim)

    # Batch_first = True
    query = torch.rand(batch_size, tgt_seq_len, embed_dim, dtype=torch.complex64)
    key = torch.rand(batch_size, src_seq_len, kdim, dtype=torch.complex64)
    value = torch.rand(batch_size, src_seq_len, vdim, dtype=torch.complex64)

    multihead_attn = c_nn.MultiheadAttention(
        embed_dim=embed_dim, num_heads=nheads, kdim=kdim, vdim=vdim, batch_first=True
    )

    attn_output, attn_output_weights = multihead_attn(query, key, value)

    assert attn_output.shape == (batch_size, tgt_seq_len, embed_dim)


def test_transformer_encoder_layer():
    nhead = 8
    seq_len = 10
    batch_size = 32
    num_features = 512

    encoder_layer = c_nn.TransformerEncoderLayer(d_model=num_features, nhead=nhead)
    src = torch.rand(seq_len, batch_size, num_features, dtype=torch.complex64)
    out = encoder_layer(src)

    assert out.shape == (seq_len, batch_size, num_features)

    encoder_layer = c_nn.TransformerEncoderLayer(
        d_model=num_features, nhead=nhead, batch_first=True
    )
    src = torch.rand(batch_size, seq_len, num_features, dtype=torch.complex64)
    out = encoder_layer(src)

    assert out.shape == (batch_size, seq_len, num_features)


def test_transformer_encoder():
    nhead = 8
    seq_len = 10
    batch_size = 32
    num_features = 512

    encoder_layer = c_nn.TransformerEncoderLayer(d_model=num_features, nhead=nhead)
    transformer_encoder = TransformerEncoder(encoder_layer, num_layers=4)

    src = torch.randn((seq_len, batch_size, num_features), dtype=torch.complex64)
    out = transformer_encoder(src)
    assert out.shape == (seq_len, batch_size, num_features)


def test_transformer_decoder_layer():
    nhead = 8
    src_seq_len = 10
    tgt_seq_len = 20
    batch_size = 32
    num_features = 512

    decoder_layer = c_nn.TransformerDecoderLayer(d_model=num_features, nhead=nhead)
    memory = torch.rand(src_seq_len, batch_size, num_features, dtype=torch.complex64)
    tgt = torch.rand(tgt_seq_len, batch_size, num_features, dtype=torch.complex64)
    out = decoder_layer(tgt, memory)

    assert out.shape == (tgt_seq_len, batch_size, num_features)

    decoder_layer = c_nn.TransformerDecoderLayer(
        d_model=num_features, nhead=nhead, batch_first=True
    )
    memory = torch.rand(batch_size, src_seq_len, num_features, dtype=torch.complex64)
    tgt = torch.rand(batch_size, tgt_seq_len, num_features, dtype=torch.complex64)
    out = decoder_layer(tgt, memory)

    assert out.shape == (batch_size, tgt_seq_len, num_features)


def test_transformer_decoder():
    nhead = 8
    src_seq_len = 10
    tgt_seq_len = 20
    batch_size = 32
    num_features = 512

    decoder_layer = c_nn.TransformerDecoderLayer(d_model=num_features, nhead=nhead)
    transformer_decoder = TransformerDecoder(decoder_layer, num_layers=4)
    memory = torch.rand((src_seq_len, batch_size, num_features), dtype=torch.complex64)
    tgt = torch.rand((tgt_seq_len, batch_size, num_features), dtype=torch.complex64)
    out = transformer_decoder(tgt, memory)

    assert out.shape == (tgt_seq_len, batch_size, num_features)


def test_transformer():
    nhead = 8
    src_seq_len = 10
    tgt_seq_len = 20
    batch_size = 32
    num_features = 512

    transformer = c_nn.Transformer(d_model=num_features, nhead=nhead, batch_first=False)

    src = torch.rand((src_seq_len, batch_size, num_features), dtype=torch.complex64)
    tgt = torch.rand((tgt_seq_len, batch_size, num_features), dtype=torch.complex64)
    out = transformer(src, tgt)

    assert out.shape == (tgt_seq_len, batch_size, num_features)

    transformer = c_nn.Transformer(d_model=num_features, nhead=nhead, batch_first=True)

    src = torch.rand((batch_size, src_seq_len, num_features), dtype=torch.complex64)
    tgt = torch.rand((batch_size, tgt_seq_len, num_features), dtype=torch.complex64)
    out = transformer(src, tgt)

    assert out.shape == (batch_size, tgt_seq_len, num_features)


if __name__ == "__main__":
    test_multihead_scaleddotproduct_selfattention()
    test_multihead_scaleddotproduct()
    # test_multihead_masks()
    test_transformer_encoder_layer()
    test_transformer_encoder()
    test_transformer_decoder_layer()
    test_transformer_decoder()
    test_transformer()



# MIT License

# Copyright (c) 2024 Jeremy Fix

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Scripts to read and parse ALOS-2 data files

The format is described in
https://www.eorc.jaxa.jp/ALOS/en/alos-2/pdf/product_format_description/PALSAR-2_xx_Format_CEOS_E_g.pdf

See also the handbook : https://www.eorc.jaxa.jp/ALOS/en/doc/alos_userhb_en.pdf
for example p99 for the image line formats

We have for High-sensitive/Fine Mode Full (Quad.) Polarimetry :
    IMG-{HH, HV, VH, VV}xxxxx__A
    LED-xxxx.1__A : SAR Leader file
    TRL-xxxx.1__A : SAR Trailer file
    VOL-xxxx.1__a : Volume directory file

    With ALOS2044980750-150324-HBQR1 as Sene ID - Product ID
    SceneID : ALOS2044980750-150324
        which stands for ALOS2 satelitte
                         04498 : orbit accumulation number
                         0750  : scene frame number
                         150324 : 2015/03/24
                            
    ProductID : HBQR1.1__A
        HBQ : High-sensitive mode Full (Quad.) polarimetry
          R : Right looking
          1.1 : Processing level
          _ : processing option not specified
          _ : map projection not specified
          A : Ascending orbit direction

If you have different settings for the polarization, the structure of the files is described in Figure 3.1-5 p29

"""

from .vol_file import VolFile
from .trailer_file import TrailerFile
from .leader_file import LeaderFile
from .sar_image import SARImage
from .dataset import ALOSDataset
